@@ -1,7 +1,38 @@
 /*
  * LTSP Graphical GTK Greeter
  * Copyright (2007) Oliver Grawert <ogra@ubuntu.com>, Canonical Ltd.
- * Code Licensed under GPL v2
+
+ * Author: Oliver Grawert <ogra@canonical.com>
+
+ * 2007, Oliver Grawert <ogra@canonical.com>
+ *       Scott Balneaves <sbalneav@ltsp.org>
+ * 2008, Ryan Niebur <RyanRyan52@gmail.com>
+ *       Warren Togami <wtogami@redhat.com>
+ *       Stéphane Graber <stgraber@ubuntu.com>
+ *       Scott Balneaves <sbalneav@ltsp.org>
+ *       Vagrant Cascadian <vagrant@freegeek.org>
+ *       Oliver Grawert <ogra@canonical.com>
+ *       John Ellson <john.ellson@comcast.net>
+ *       Gideon Romm <gadi@ltsp.org>
+ *       Jigish Gohil <cyberorg@opensuse.org>
+ *       Wolfgang Schweer <schweer@cityweb.de>
+ *       Toshio Kuratomi
+
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation; version 2.
+
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+
+ * You should have received a copy of the GNU General Public License
+ * along with this program.  If not, you can find it on the World Wide
+ * Web at http://www.gnu.org/copyleft/gpl.html, or write to the Free
+ * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+ * MA 02110-1301, USA.
+
  */
 
 #define _GNU_SOURCE
@@ -17,6 +48,7 @@
 #include <net/if.h>
 #include <arpa/inet.h>
 #include <sys/ioctl.h>
+#include <gdk/gdkkeysyms.h> 
 
 #include <greeter.h>
 
@@ -33,25 +65,24 @@
 GtkWidget *UserPrompt;          /* prompt area before the entry */
 GtkWidget *StatusMessages;      /* Status msg area below entry */
 GtkWidget *entry;               /* entry box */
-GtkWidget *guestbox;
+
+GtkWidget *GuestButton;
 GtkWidget *timeoutbox;
 gboolean timeout_enabled;
 
 GHashTable *ldminfo_hash = NULL;
 GList *host_list = NULL;
 GIOChannel *g_stdout;           /* stdout io channel */
-char ldm_theme_dir[512];
+gchar * ldm_theme_dir;
 
 int allowguest;
+gint login_timeout;
 gint timeout_left;
 
-char *
+gchar *
 ldm_theme_file(char * file)
 {
-    static char path[512];
-    strcpy(path, ldm_theme_dir);
-    strcat(path, file);
-    return path;
+    return g_strconcat(ldm_theme_dir, file, NULL);
 }
 
 static void
@@ -64,14 +95,14 @@ static void
 halt(GtkWidget * widget, gpointer data)
 {
     GError **error = NULL;
-    g_spawn_command_line_async("/sbin/poweroff -fp", error);
+    g_spawn_command_line_async("/sbin/poweroff -p", error);
 }
 
 static void
 reboot(GtkWidget * widget, gpointer data)
 {
     GError **error = NULL;
-    g_spawn_command_line_async("/sbin/reboot -fp", error);
+    g_spawn_command_line_async("/sbin/reboot -p", error);
 }
 
 gboolean
@@ -86,12 +117,12 @@ update_time(GtkWidget * label)
 
     // use 12 hour clock format if LDM_12HOURCLOCK is set to true
     if (ldm_getenv_bool("LDM_12HOURCLOCK")) {
-        timestring = g_strdup_printf("%.2d:%.2d ",
+        timestring = g_strdup_printf("<b>%.2d:%.2d</b>",
                                    (timePtr->tm_hour % 12) ? (timePtr->tm_hour % 12) : 12,
                                    timePtr->tm_min);
     }
     else {
-        timestring = g_strdup_printf("%.2d:%.2d ",
+        timestring = g_strdup_printf("<b>%.2d:%.2d</b>",
                                    timePtr->tm_hour, timePtr->tm_min);
     }
 
@@ -117,7 +148,7 @@ update_timeout(GtkWidget * label)
             timeout_left = 0;
             timeout_enabled = FALSE;
         } else if (timeout_left == 0){
-            timeout_left = atoi(getenv("LDM_TIMEOUT_TIME"));
+            timeout_left = login_timeout;
         }
         string = g_strdup_printf(_("Automatic login in %d seconds"), timeout_left);
         gtk_label_set_markup((GtkLabel *) label, string);
@@ -127,6 +158,7 @@ update_timeout(GtkWidget * label)
         timeout_left = 0;
         gtk_widget_hide(timeoutbox);
     }
+    return TRUE;
 }
 
 void
@@ -136,6 +168,29 @@ destroy_popup(GtkWidget * widget, GtkWidget * popup)
     return;
 }
 
+gboolean
+xproperty_exists(gchar * property)
+{
+    GdkAtom *actual_property_type;
+    gint *actual_format;
+    gint *actual_length;
+    guchar **data;
+    gboolean result;
+
+    result=gdk_property_get(
+        gdk_get_default_root_window(),
+        gdk_atom_intern(property,TRUE),
+        GDK_NONE,
+        0,
+        512,
+        FALSE,
+        actual_property_type,
+        actual_format,
+        actual_length,
+        data);
+    return result;
+}
+
 gboolean
 handle_command(GIOChannel * io_input)
 {
@@ -150,11 +205,12 @@ handle_command(GIOChannel * io_input)
         split_buf = g_strsplit(buf->str, " ", 2);
         gtk_label_set_markup((GtkLabel *) StatusMessages, split_buf[1]);
         g_strfreev(split_buf);
-        gtk_entry_set_editable(GTK_ENTRY(entry), FALSE);
     } else if (!g_strncasecmp(buf->str, "quit", 4)) {
-        GdkCursor *cursor;
-        cursor = gdk_cursor_new(GDK_WATCH);
-        gdk_window_set_cursor(gdk_get_default_root_window(), cursor);
+        if (!xproperty_exists("X11VNC_TICKER")) {
+            GdkCursor *cursor;
+            cursor = gdk_cursor_new(GDK_WATCH);
+            gdk_window_set_cursor(gdk_get_default_root_window(), cursor);
+        }
         gtk_main_quit();
     } else if (!g_strncasecmp(buf->str, "prompt", 6)) {
         gchar **split_buf;
@@ -162,19 +218,22 @@ handle_command(GIOChannel * io_input)
         gtk_label_set_markup((GtkLabel *) UserPrompt, split_buf[1]);
         g_strfreev(split_buf);
     } else if (!g_strncasecmp(buf->str, "userid", 6)) {
-        gtk_widget_show(guestbox);
-        timeout_enabled = ldm_getenv_bool("LDM_TIMEOUT_ENABLED");
+        timeout_enabled = login_timeout > 0;
         if(timeout_enabled) {
             gtk_widget_show(timeoutbox);
         }
+        gtk_entry_set_text(entry, "");
         gtk_entry_set_visibility(GTK_ENTRY(entry), TRUE);
-        gtk_entry_set_editable(GTK_ENTRY(entry), TRUE);
+        gtk_widget_set_sensitive(GTK_WIDGET(entry), TRUE);
+        gtk_widget_set_sensitive(GTK_WIDGET(GuestButton), TRUE);
     } else if (!g_strncasecmp(buf->str, "passwd", 6)) {
         timeout_enabled = FALSE;
         timeout_left = 0;
         gtk_widget_hide(timeoutbox);
+        gtk_entry_set_text(entry, "");
         gtk_entry_set_visibility(GTK_ENTRY(entry), FALSE);
-        gtk_entry_set_editable(GTK_ENTRY(entry), TRUE);
+        gtk_widget_set_sensitive(GTK_WIDGET(entry), TRUE);
+        gtk_widget_set_sensitive(GTK_WIDGET(GuestButton), FALSE);
     } else if (!g_strncasecmp(buf->str, "hostname", 8)) {
         gchar *hoststr;
         update_selected_host();
@@ -182,6 +241,8 @@ handle_command(GIOChannel * io_input)
         g_io_channel_write_chars(g_stdout, hoststr, -1, NULL, NULL);
         g_io_channel_flush(g_stdout, NULL);
         g_free(hoststr);
+        gtk_widget_set_sensitive(GTK_WIDGET(entry), FALSE);
+        gtk_widget_set_sensitive(GTK_WIDGET(GuestButton), FALSE);
     } else if (!g_strncasecmp(buf->str, "language", 8)) {
         gchar *langstr;
         update_selected_lang();
@@ -233,7 +294,8 @@ handle_entry(GtkEntry * entry, GdkWindow * window)
     g_io_channel_write_chars(g_stdout, entrystr, -1, NULL, NULL);
     g_io_channel_flush(g_stdout, NULL);
     g_free(entrystr);
-    gtk_entry_set_text(entry, "");
+    if(gtk_entry_get_visibility(GTK_ENTRY(entry)))
+        gtk_entry_set_text(entry, "");
 }
 
 static void
@@ -319,51 +381,21 @@ scopy(char *dest, char *source)
     return dest;
 }
 
-void
-setup_keyboard_layout()
+/*
+ * Remap GDK_Tab to behaves as enter
+ * Inspired by gdm/gui/gdmlogin.c
+ */
+static gboolean
+key_press_event( GtkWidget *entry, GdkEventKey *event, gpointer window)
 {
-    char *argv[32];
-    char layout[MAXSTRSZ];
-    char model[MAXSTRSZ];
-    char rules[MAXSTRSZ];
-    char options[MAXSTRSZ];
-    char variant[MAXSTRSZ];
-    int i = 0;
-    int pid;
+    if ((event->keyval == GDK_Tab ||
+         event->keyval == GDK_KP_Tab) &&
+        (event->state & (GDK_CONTROL_MASK|GDK_MOD1_MASK|GDK_SHIFT_MASK)) == 0) {
+            handle_entry(GTK_ENTRY(entry), window);
+            return TRUE;
+    }
 
-    argv[i++] = "/usr/bin/setxkbmap";
-    scopy(layout, getenv("XKBLAYOUT"));
-    if (! layout[0] == '\0') {
-        argv[i++] = "-layout";
-        argv[i++] = layout;
-    }
-    scopy(model, getenv("XKBMODEL"));
-    if (! model[0] == '\0') {
-        argv[i++] = "-model";
-        argv[i++] = model;
-    }
-    scopy(rules, getenv("XKBRULES"));
-    if (! rules[0] == '\0') {
-        argv[i++] = "-rules";
-        argv[i++] = rules;
-    }
-    scopy(options, getenv("XKBOPTIONS"));
-    if (! options[0] == '\0') {
-        argv[i++] = "-option";
-        argv[i++] = options;
-    }
-    scopy(variant, getenv("XKBVARIANT"));
-    if (! variant[0] == '\0') {
-        argv[i++] = "-variant";
-        argv[i++] = variant;
-    }
-    argv[i++] = NULL;
-
-    // Skip setxkbmap if no parameters were specified
-    if (i > 2) {
-        g_spawn_sync(NULL,argv,NULL,G_SPAWN_STDOUT_TO_DEV_NULL|G_SPAWN_STDERR_TO_DEV_NULL,NULL,NULL,NULL,NULL,NULL,NULL);
-    }
-    return;
+    return FALSE;
 }
 
 int
@@ -375,8 +407,8 @@ main(int argc, char *argv[])
     gint lw, lh;
 
     GdkCursor *normcursor, *busycursor;
-    GtkWidget *window, *syslabel, *logo, *EntryBox, *timelabel;
-    GtkWidget *GuestButton, *StatusBarBox, *spacer, *guestspacer1, *guestspacer2, *vbox, *vbox2, *hbox;
+    GtkWidget *window, *syslabel, *logo, *EntryBox, *timelabel, *bottom_right;
+    GtkWidget *guestbox, *BottomBarBox, *spacer, *guestspacer1, *guestspacer2, *vbox, *vbox2, *hbox, *TopBarBox;
     GtkWidget *timeoutspacer1, *timeoutspacer2, *timeoutlabel;
     GtkWidget *entryspacer1, *entryspacer2;
     GtkButton *optionbutton, *cancelbutton;
@@ -384,9 +416,10 @@ main(int argc, char *argv[])
     GdkPixbuf *rawpic, *pix;
     GdkPixmap *pic;
     GdkBitmap *mask;
-    gint width, height;
+    gint width, height, has_bottom_right_image;
     GIOChannel *g_stdin;
-
+    GtkWidget *optionbutton_box, *optionbutton_spacer;
+    GtkWidget *bottom_right_box, *bottom_right_spacer;
     char * ldm_theme;
 
 #ifdef ENABLE_NLS
@@ -397,22 +430,18 @@ main(int argc, char *argv[])
 #endif
 
     gtk_init(&argc, &argv);
-
-    *ldm_theme_dir='\0';
     ldm_theme = getenv("LDM_THEME");
 
     if (ldm_theme)
     {
-        char temp[512];
-        strcpy(temp, ldm_theme);
-        if(*temp != '/')
-            strcpy(ldm_theme_dir, LDM_THEME_DIR);
-        strcat(ldm_theme_dir, temp);
+        if(*ldm_theme == '/')
+            ldm_theme_dir = g_strdup(ldm_theme);
+        else
+            ldm_theme_dir = g_strconcat(LDM_THEME_DIR, ldm_theme, NULL);
     }
     else
     {
-        strcpy(ldm_theme_dir, LDM_THEME_DIR);
-        strcat(ldm_theme_dir, "default");
+        ldm_theme_dir = g_strconcat(LDM_THEME_DIR, "default", NULL);
     }
 
     allowguest = ldm_getenv_bool("LDM_GUESTLOGIN");
@@ -439,6 +468,14 @@ main(int argc, char *argv[])
     h = height;
 
     logo = gtk_image_new_from_file(ldm_theme_file("/logo.png"));
+#ifdef K12LINUX
+    if(access(ldm_theme_file("/bottom_right.png"), R_OK) == 0) {
+        bottom_right = gtk_image_new_from_file(ldm_theme_file("/bottom_right.png"));
+        has_bottom_right_image=TRUE;
+    }
+    else
+        has_bottom_right_image=FALSE;
+#endif
 
     pix = gtk_image_get_pixbuf((GtkImage *) logo);
     lw = gdk_pixbuf_get_width(pix);
@@ -462,7 +499,12 @@ main(int argc, char *argv[])
     EntryBox = gtk_hbox_new(FALSE, 5);
     hbox = gtk_hbox_new(FALSE, 0);
 
-    StatusBarBox = gtk_hbox_new(FALSE, 0);
+#ifndef K12LINUX
+    BottomBarBox = gtk_hbox_new(FALSE, 0);
+#else
+    BottomBarBox = gtk_hbox_new(FALSE, 0);
+    TopBarBox = gtk_hbox_new(FALSE, 5);
+#endif
 
     optionbutton =
         (GtkButton *) gtk_button_new_from_stock("gtk-preferences");
@@ -486,21 +528,51 @@ main(int argc, char *argv[])
 
     syslabel = gtk_label_new("");
     timelabel = gtk_label_new("");
+#ifdef K12LINUX
     hoststring =
-        g_strdup_printf("<b>%s (%s) //</b>", get_sysname(), getenv("LDMINFO_IPADDR"));
+        g_strdup_printf("<b>%s</b> (%s)", get_sysname(), getenv("LDMINFO_IPADDR"));
+#else
+    hoststring =
+        g_strdup_printf("<b>%s</b> (%s) //", get_sysname(), getenv("LDMINFO_IPADDR"));
+#endif
     gtk_label_set_markup((GtkLabel *) syslabel, hoststring);
     update_time(timelabel);
 
     g_timeout_add(30000, (GSourceFunc) update_time, timelabel);
 
-    gtk_box_pack_start(GTK_BOX(StatusBarBox),
+#ifndef K12LINUX
+    gtk_box_pack_start(GTK_BOX(BottomBarBox),
                        GTK_WIDGET(optionbutton), FALSE, FALSE, 5);
-    /*gtk_box_pack_start(GTK_BOX(StatusBarBox),
-                       GTK_WIDGET(cancelbutton), FALSE, FALSE, 5);*/
-    gtk_box_pack_end(GTK_BOX(StatusBarBox),
+#else
+    optionbutton_box = gtk_vbox_new(FALSE, 0);
+    optionbutton_spacer = gtk_label_new("");
+    gtk_box_pack_start(GTK_BOX(optionbutton_box), GTK_WIDGET(optionbutton_spacer), TRUE, FALSE, 0);
+    gtk_box_pack_end(GTK_BOX(optionbutton_box), GTK_WIDGET(optionbutton), FALSE, FALSE, 5);
+    gtk_box_pack_start(GTK_BOX(BottomBarBox),
+                       GTK_WIDGET(optionbutton_box), FALSE, FALSE, 5);
+#endif
+
+#ifndef K12LINUX
+    gtk_box_pack_end(GTK_BOX(BottomBarBox),
                      GTK_WIDGET(timelabel), FALSE, FALSE, 5);
-    gtk_box_pack_end(GTK_BOX(StatusBarBox),
+#else
+    if(has_bottom_right_image == TRUE) {
+        bottom_right_box = gtk_vbox_new(FALSE, 0);
+        bottom_right_spacer = gtk_label_new("");
+        gtk_box_pack_start(GTK_BOX(bottom_right_box), GTK_WIDGET(bottom_right_spacer), TRUE, FALSE, 0);
+        gtk_box_pack_end(GTK_BOX(bottom_right_box), GTK_WIDGET(bottom_right), FALSE, FALSE, 0);
+        gtk_box_pack_end(GTK_BOX(BottomBarBox),
+                         GTK_WIDGET(bottom_right_box), FALSE, FALSE, 0);
+    }
+    gtk_box_pack_start(GTK_BOX(TopBarBox),
+                     GTK_WIDGET(syslabel), FALSE, FALSE, 5);
+    gtk_box_pack_end(GTK_BOX(TopBarBox),
+                     GTK_WIDGET(timelabel), FALSE, FALSE, 5);
+#endif
+#ifndef K12LINUX
+    gtk_box_pack_end(GTK_BOX(BottomBarBox),
                      GTK_WIDGET(syslabel), FALSE, FALSE, 0);
+#endif
 
     UserPrompt = gtk_label_new("");
     spacer = gtk_label_new("");
@@ -516,9 +588,17 @@ main(int argc, char *argv[])
     gtk_entry_set_width_chars(GTK_ENTRY(entry), 12);
     g_signal_connect(G_OBJECT(entry), "activate",
                      G_CALLBACK(handle_entry), root);
+    // Remap tab on entry to behave as enter
+    g_signal_connect(G_OBJECT(entry), "key_press_event",
+                     G_CALLBACK(key_press_event), window);
 
-    timeout_enabled = ldm_getenv_bool("LDM_TIMEOUT_ENABLED");
-    timeout_left = 0;
+
+    if (getenv("LDM_LOGIN_TIMEOUT") != NULL ) {
+        login_timeout = atoi(getenv("LDM_LOGIN_TIMEOUT"));
+    } else {
+        login_timeout = 0;
+    }
+    timeout_enabled = login_timeout > 0;
     timeout_left = 0;
     timeoutspacer1 = gtk_label_new("");
     timeoutspacer2 = gtk_label_new("");
@@ -535,7 +615,6 @@ main(int argc, char *argv[])
     GuestButton = gtk_button_new_with_label (_("Login as Guest"));
     g_signal_connect(G_OBJECT(GuestButton), "clicked",
         G_CALLBACK(handle_guestbutton), root);
-    gtk_button_set_relief((GtkButton *)GuestButton, GTK_RELIEF_NONE);
     gtk_button_set_focus_on_click((GtkButton *)GuestButton, FALSE);
     guestbox = gtk_hbox_new(FALSE, 0);
     gtk_box_pack_start(GTK_BOX(guestbox), guestspacer1, TRUE, FALSE, 0);
@@ -549,6 +628,17 @@ main(int argc, char *argv[])
     gtk_box_pack_start(GTK_BOX(EntryBox), entry, FALSE, FALSE, 0);
     gtk_box_pack_start(GTK_BOX(EntryBox), entryspacer2, TRUE, FALSE, 0);
 
+#ifdef K12LINUX
+    gtk_box_pack_start(GTK_BOX(vbox2), TopBarBox, FALSE, FALSE, 5);
+
+    gtk_container_add(GTK_CONTAINER(window), vbox2);
+
+    gtk_widget_show_all(window);
+
+    gdk_window_set_cursor(root, normcursor);
+
+    spheight -= TopBarBox->allocation.height;
+#endif
     gtk_box_pack_start(GTK_BOX(vbox), spacer, FALSE, FALSE, spheight);
     gtk_box_pack_start(GTK_BOX(vbox), logo, FALSE, FALSE, 5);
     gtk_box_pack_start(GTK_BOX(vbox), EntryBox, TRUE, FALSE, 0);
@@ -557,13 +647,21 @@ main(int argc, char *argv[])
     gtk_box_pack_start(GTK_BOX(vbox), StatusMessages, TRUE, FALSE, 0);
     gtk_box_pack_start(GTK_BOX(hbox), vbox, TRUE, FALSE, 0);
     gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, FALSE, 0);
-    gtk_box_pack_end(GTK_BOX(vbox2), StatusBarBox, FALSE, FALSE, 5);
+#ifndef K12LINUX
+    gtk_box_pack_end(GTK_BOX(vbox2), BottomBarBox, FALSE, FALSE, 5);
 
     gtk_container_add(GTK_CONTAINER(window), vbox2);
+#else
+    gtk_box_pack_end(GTK_BOX(vbox2), BottomBarBox, FALSE, FALSE, 0);
+#endif
 
     gtk_widget_show_all(window);
 
+#ifndef K12LINUX
     gdk_window_set_cursor(root, normcursor);
+#else
+    gtk_widget_grab_focus(GTK_WIDGET(entry));
+#endif
 
     /*
      * Start listening to stdin
@@ -573,7 +671,6 @@ main(int argc, char *argv[])
     g_stdout = g_io_channel_unix_new(STDOUT_FILENO);
     g_io_add_watch(g_stdin, G_IO_IN, (GIOFunc) handle_command, g_stdin);
 
-    setup_keyboard_layout();
     gtk_main();
 
     return 0;
