@@ -3,16 +3,45 @@
  * LTSP display manager.
  * Manages spawning a session to a server.
  *
- * (c) Scott Balneaves, sbalneav@ltsp.org
+ * Copyright 2007, Scott Balneaves <sbalneav@ltsp.org>
  *
- * This software is licensed under the GPL v2 or later.
+
+ * Author: Scott Balneaves <sbalneav@ltsp.org>
+
+ * 2007, Scott Balneaves <sbalneav@ltsp.org>
+ *       Oliver Grawert <ogra@canonical.com>
+ *       Vagrant Cascadian <vagrant@freegeek.org>
+ * 2008, Scott Balneaves <sbalneav@ltsp.org>
+ *       Ryan Niebur <RyanRyan52@gmail.com>
+ *       John Ellson <john.ellson@comcast.net>
+ *       Gideon Romm <gadi@ltsp.org>
+ *       Warren Togami <wtogami@redhat.com>
+ *       Vagrant Cascadian <vagrant@freegeek.org>
+ *       Francis Giraldeau <francis.giraldeau@revolutionlinux.com>
+ *       Toshio Kuratomi
+
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation; either version 2 of the
+ * License, or (at your option) any later version.
+
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+
+ * You should have received a copy of the GNU General Public License
+ * along with this program.  If not, you can find it on the World Wide
+ * Web at http://www.gnu.org/copyleft/gpl.html, or write to the Free
+ * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+ * MA 02110-1301, USA.
+
  */
 
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdarg.h>
 #include <unistd.h>
-#include <syslog.h>
 #include <string.h>
 #include <sys/wait.h>
 #include <glib.h>
@@ -21,7 +50,7 @@
 #include "ldm.h"
 
 int
-get_userid()
+get_userid(void)
 {
     gchar *prompt;
     gchar *isguest;
@@ -63,7 +92,7 @@ get_userid()
 }
 
 int
-get_passwd()
+get_passwd(void)
 {
     gchar *prompt;
     gchar *command = "passwd\n";
@@ -98,7 +127,7 @@ set_message(gchar *message)
     gchar *prompt;
 
     if (ldm.autologin) {
-        return;
+        return 0;
     }
 
     prompt = g_strconcat("msg ", message, "\n", NULL);
@@ -115,7 +144,7 @@ set_message(gchar *message)
 }
 
 int
-get_host()
+get_host(void)
 {
     char *cmd = "hostname\n";
     gsize l, t;
@@ -137,7 +166,7 @@ get_host()
 }
 
 int
-get_language()
+get_language(void)
 {
     char *cmd = "language\n";
     gchar *lang;
@@ -166,7 +195,7 @@ get_language()
 }
 
 int
-get_session()
+get_session(void)
 {
     char *cmd = "session\n";
     gchar *session;
@@ -195,16 +224,26 @@ get_session()
 }
 
 void
-close_greeter()
+close_greeter(void)
 {
     gint rfd, wfd;
+    gboolean failed = FALSE;
     if (!ldm.greeterpid) {
         return;
     }
-
-    if (kill(ldm.greeterpid, SIGTERM) < 0) {
-        fprintf(ldmlog, "Sending SIGTERM to greeter failed\n");
-        kill(ldm.greeterpid, SIGKILL);
+    if (g_io_channel_write_chars(ldm.greeterw, "quit\n", -1, NULL, NULL) == G_IO_STATUS_NORMAL) {
+        if(g_io_channel_flush(ldm.greeterw, NULL) != G_IO_STATUS_NORMAL) {
+            failed = TRUE;
+        }
+    } else {
+        failed = TRUE;
+    }
+    if(failed == TRUE) {
+        logerr("Sending greeter the quit command failed");
+        if (kill(ldm.greeterpid, SIGTERM) < 0) {
+            logerr("Sending SIGTERM to greeter failed");
+            kill(ldm.greeterpid, SIGKILL);
+        }
     }
     ldm_wait(ldm.greeterpid);
     rfd = g_io_channel_unix_get_fd(ldm.greeterr);
