@@ -3,15 +3,29 @@
  * LTSP display manager.
  * Manages spawning a session to a server.
  *
- * (c) Scott Balneaves, sbalneav@ltsp.org,      2007, 2008 
+ * Copyright Scott Balneaves, sbalneav@ltsp.org, 2007, 2008
  *     Oliver Grawert, ogra@ubuntu.com,         2005, 2006, 2007
- *     Vagrant Cascadian, vagrant@freegeek.org, 
+ *     Vagrant Cascadian, vagrant@freegeek.org, 2008
  *     Gideon Romm, gadi@ltsp.org,              2007, 2008
- *     Warren Togami
+ *     Warren Togami, wtogami@redhat.com        2008
  *     Ryan Niebur, RyanRyan52@gmail.com,       2008
- *
- *
- * This software is licensed under the GPL v2 or later.
+
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation; either version 2 of the
+ * License, or (at your option) any later version.
+
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+
+ * You should have received a copy of the GNU General Public License
+ * along with this program.  If not, you can find it on the World Wide
+ * Web at http://www.gnu.org/copyleft/gpl.html, or write to the Free
+ * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+ * MA 02110-1301, USA.
+
  */
 
 #include <stdio.h>
@@ -19,6 +33,7 @@
 #include <unistd.h>
 #include <syslog.h>
 #include <string.h>
+#include <errno.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
@@ -33,16 +48,56 @@
 #include <locale.h>
 #define _(text) gettext(text)
 
+#define LOGFILE "/var/log/ldm.log"
+
+GHashTable *ldminfo_hash = NULL;
+GList *host_list = NULL;
+volatile sig_atomic_t unexpected_child = 0;
+volatile sig_atomic_t child_exited = 0;
+FILE *logfile = NULL;
 
 struct ldm_info ldm;
-int    logfd;
-FILE  *ldmlog;
+
+/*
+ * loginfo: logs info messages to either syslog or the log file.
+ */
 
 void
-usage()
+loginfo(const char *format, ...)
 {
-    fprintf(stderr, "Usage: ldm <vt[1-N]> <:[0-N]> [xserver options]\n");
-    exit(1);
+    va_list ap;
+
+    va_start(ap, format);
+
+    if (logfile) {
+        vfprintf(logfile, format, ap);
+        fprintf(logfile, "\n");
+    } else {
+        vsyslog(LOG_INFO, format, ap);
+    }
+
+    va_end(ap);
+}
+
+/*
+ * logerr: logs error messages to either syslog or the log file.
+ */
+
+void
+logerr(const char *format, ...)
+{
+    va_list ap;
+
+    va_start(ap, format);
+
+    if (logfile) {
+        vfprintf(logfile, format, ap);
+        fprintf(logfile, "\n");
+    } else {
+        vsyslog(LOG_ERR, format, ap);
+    }
+
+    va_end(ap);
 }
 
 /*
@@ -54,7 +109,7 @@ usage()
 void
 die(char *msg)
 {
-    fprintf(ldmlog, "%s\n", msg);
+    logerr("%s", msg);
 
     /*
      * Shut things down gracefully if we can
@@ -63,16 +118,16 @@ die(char *msg)
     if (ldm.greeterpid) {
         close_greeter();
     }
-    if (ldm.sshpid) {
-        ssh_endsession();
-    }
-    if (ldm.xserverpid) {
-        kill(ldm.xserverpid, SIGTERM);      /* Kill Xorg server off */
-        ldm_wait(ldm.xserverpid);
+
+    ssh_endsession();
+
+    /* Stop logging */
+    if (logfile) {
+        fclose(logfile);
+    } else {
+        closelog();
     }
 
-    fclose(ldmlog);
-    close(logfd);
     exit(1);
 }
 
@@ -83,8 +138,7 @@ die(char *msg)
  */
 
 GPid
-ldm_spawn (gchar *command, gint *rfd, gint *wfd,  GSpawnChildSetupFunc setup,
-           gpointer data)
+ldm_spawn (gchar *command, gint *rfd, gint *wfd,  GSpawnChildSetupFunc setup)
 {
     GPid pid;
     GError *error = NULL;
@@ -92,88 +146,94 @@ ldm_spawn (gchar *command, gint *rfd, gint *wfd,  GSpawnChildSetupFunc setup,
     gint argc;
     gchar **argv = NULL;
 
-    g_shell_parse_argv(command, &argc, &argv, NULL);
+    g_shell_parse_argv (command, &argc, &argv, NULL);
 
     if (!wfd) {
         flags |= G_SPAWN_STDOUT_TO_DEV_NULL;
     }
 
-    g_spawn_async_with_pipes(
+    g_spawn_async_with_pipes (
                   NULL,             /* Working directory: inherit */
                   argv,             /* Arguments, null term */
                   NULL,             /* Environment, inherit from parent */
                   flags,            /* Flags, set above */
                   setup,            /* Child setup function: passed to us */
-                  data,             /* user_data, passed to us */
+                  NULL,             /* No user data */
                   &pid,             /* child pid */
                   wfd,              /* Pointer to in file descriptor */
                   rfd,              /* Pointer to out file descriptor */
-                  &logfd,           /* Our logfile file descriptor */
+                  NULL,             /* No stderr */
                   &error);          /* GError handler */
 
     g_strfreev(argv);
 
     if (error) {
-        fprintf(ldmlog, _("ldm_spawn failed to execute: %s\n"), error->message);
+        logerr(_("ldm_spawn failed to execute: %s"), error->message);
         die(_("Exiting ldm"));
+    } else {
+        loginfo(_("ldm_spawn: pid = %d"), pid);
     }
 
     return pid;
 }
 
+/*
+ * handle_sigchld
+ *
+ * Handle sigchld's for ldm processes.  Empty function,
+ * since we wait for things to happen in order via
+ * ldm_wait
+ */
+
+void
+handle_sigchld(int signo)
+{
+    /* do nothing */
+    child_exited = TRUE;
+}
 
 /*
  * ldm_wait
  *
- * Reap child process
- */
-
-int
-ldm_wait(GPid pid)
-{
-    int status;
-
-    if (waitpid (pid, &status, 0) < 0) {
-        die(_("Error: wait() call failed"));
-    }
-    if (!WIFEXITED (status)) {
-        fprintf(ldmlog, _("Process returned no status\n"));
-        return 1;
-    } else {
-        return WEXITSTATUS (status);
-    }
-}
-
-/*
- * launch_x
- *
- * Launch the Xorg server
+ * wait for child process
  */
 
 void
-launch_x(gchar **xserver_args)
+ldm_wait(GPid pid)
 {
-    gchar *command;
-    gchar **ptr = xserver_args;
-    gchar *server_args = NULL;
+    siginfo_t info;
+    do {
+        int res;
+        res = waitid (P_ALL, 0, &info, WEXITED | WSTOPPED);
+        if (res == -1) {
+            int temp;
+            temp = errno;
+            logerr(g_strjoin(" ", _("waitid returned an error:"), strerror(errno), NULL));
+            if(temp == ECHILD) {
+                break;
+            }
+        } else {
+            if (info.si_pid == pid) {
+                /*
+                 * The process we were waiting for exited,
+                 * so break out of the loop.
+                 */
+                break;
+            } else {
+                logerr(_("Unexpected terminated process: pid = %d"),
+                            info.si_pid);
+                unexpected_child = TRUE;
+            }
+        }
+    } while (TRUE);
 
-    ptr += 3;
-
-    if (*ptr) {
-        server_args = g_strjoinv(" ", ptr);
-    } else {
-        server_args = g_strdup("");
+    if (info.si_code == CLD_EXITED) {
+        loginfo(_("Process %d exited with status %d"), info.si_pid,
+                WEXITSTATUS(info.si_status));
+    } else if (info.si_code == CLD_KILLED) {
+        logerr(_("Process %d killed by signal %d"), info.si_pid,
+                info.si_status);
     }
-
-    /* Spawn the X server */
-    command = g_strjoin(" ", "X", "-auth", ldm.authfile, "-br", "-noreset",
-            server_args, ldm.vty, ldm.display, NULL);
-
-    fprintf(ldmlog, _("Launching X as: %s\n"), command);
-    /* Pass logfd as stdout so we get X messages */
-    ldm.xserverpid = ldm_spawn(command, NULL, &logfd, NULL, NULL);
-    g_free(server_args);
-    g_free(command);
 }
 
 /*
@@ -190,12 +250,32 @@ rc_files(char *action)
 
     command = g_strjoin(" ", "/bin/sh",  RC_DIR "/ldm-script", action, NULL);
 
-    fprintf(ldmlog, "rc_files: %s\n", command);
-    rcpid = ldm_spawn(command, NULL, NULL, NULL, NULL);
+    loginfo(_("rc_files: %s"), command);
+    rcpid = ldm_spawn(command, NULL, NULL, NULL);
+
     ldm_wait(rcpid);
     g_free(command);
 }
 
+void
+get_Xsession(void)
+{
+    ldminfo *curr_host = NULL;
+    ldm.xsession = NULL;
+    ldm.xsession = g_strdup(getenv("LDM_XSESSION"));
+    if (!ldm.xsession || strlen(ldm.xsession) == 0) {
+        curr_host = g_hash_table_lookup(ldminfo_hash, ldm.server);
+        ldm.xsession = curr_host->xsession;
+    }
+    if (!ldm.xsession || strlen(ldm.xsession) == 0) {
+       // stupid fallback
+       ldm.xsession = g_strdup(getenv("LDM_DEFAULT_XSESSION"));
+    }
+    if (!ldm.xsession || strlen(ldm.xsession) == 0) {
+        die (g_strconcat(_("ERROR: no Xsession"), "\n", NULL ));
+    }
+}
+
 /*
  * x_session
  *
@@ -203,90 +283,16 @@ rc_files(char *action)
  */
 
 void
-x_session()
+x_session(void)
 {
-    gchar *command;
-    gchar *display = NULL;
-    gchar *sound = NULL;
-    gchar *lang = NULL;
-    gchar *localdev = NULL;
-    gchar *client_env = NULL;
-    gchar *esdcmd = "esd -nobeeps -public -tcp";
-    GPid xsessionpid;
-    GPid esdpid = 0;
+    get_Xsession();
 
-    /*
-     * Set up the LTSP_CLIENT environment variable.
-     * We may not want to force set this:
-     * FIXME:  Revisit this at some point.
-     */
-
-    client_env = g_strconcat("LTSP_CLIENT=", ldm.ipaddr, NULL);
-
-    /*
-     * Set the DISPLAY env, if not running over encrypted ssh
-     */
-
-    if (ldm.directx) {
-        display = g_strconcat("DISPLAY=", ldm.ipaddr, ldm.display, NULL);
-    }
-
-    /*
-     * Set our language, if a different one is picked.
-     */
-
-    if (ldm.lang) {
-        lang = g_strconcat("LC_ALL=", ldm.lang, " LANGUAGE=", ldm.lang,
-                           " LANG=", ldm.lang, NULL);
-    }
-
-    /*
-     * Handle sound
-     */
-
-    if (ldm.sound) {
-        if (!ldm.sound_daemon || !strncmp(ldm.sound_daemon, "pulse", 5)) {
-            sound = g_strconcat("PULSE_SERVER=tcp:", ldm.ipaddr,
-                    ":4713 ESPEAKER=", ldm.ipaddr, ":16001", NULL);
-        } else if (!strncmp(ldm.sound_daemon, "esd", 3)) {
-            sound = g_strconcat("ESPEAKER=", ldm.ipaddr, ":16001", NULL);
-            esdpid = ldm_spawn(esdcmd, NULL, NULL, NULL, NULL); /* launch ESD */
-        } else if (!strncmp(ldm.sound_daemon, "nasd", 4)) {
-            sound = g_strconcat("AUDIOSERVER=", ldm.ipaddr, ":0", NULL);
-        }
-    }
-
-    /*
-     * Handle localdev
-     */
-
-    if (ldm.localdev) {
-        localdev = "; /usr/sbin/ltspfsmounter all cleanup";
-    }
-
-    /*
-     * FIXME: ditch the kill -1 and use ssh -o exit on the control socket 
-     */
-    command = g_strjoin(" ", "ssh", "-Y", "-t", "-S", ldm.control_socket,
-            "-l", ldm.username, ldm.server, client_env,
-            lang ? lang : "", display ? display : "", sound ? sound : "",
-            ldm.xsession, ldm.session, localdev ? localdev : "",
-            "; kill -1 $PPID", NULL);
-
-    fprintf(ldmlog, _("x_session: %s\n"), command);
-    xsessionpid = ldm_spawn(command, NULL, NULL, NULL, NULL);
-    ldm_wait(xsessionpid);
-    if (esdpid) {
-        kill(esdpid, SIGTERM);
-        ldm_wait(esdpid);
-        esdpid = 0;
-    }
-    g_free(lang);
-    g_free(display);
-    g_free(sound);
-    g_free(client_env);
-    g_free(command);
+    setenv("LDM_SESSION", ldm.session, 1);
+    setenv("LDM_XSESSION", ldm.xsession, 1);
+    if(ldm.lang)
+        setenv("LDM_LANGUAGE", ldm.lang, 1);
 
+    rc_files("xsession");
 }
 
 /*
@@ -294,13 +300,15 @@ x_session()
  */
 
 void
-load_guestinfo()
+load_guestinfo(void)
 {
        char **hosts_char = NULL;
        gchar *autoservers = NULL;
+       gboolean good;
+       int i;
 
        /* Get all info for autologin, without greeter */
-       fprintf(ldmlog, _("Logging in as guest\n"));
+       loginfo(_("Logging in as guest"));
        ldm.username = g_strdup(getenv("LDM_USERNAME"));
        ldm.password = g_strdup(getenv("LDM_PASSWORD"));
        if (!ldm.username) {
@@ -320,19 +328,32 @@ load_guestinfo()
            autoservers = g_strdup(getenv("LDM_SERVER"));
        }
 
-       /*
-        * Make sure that if there are multiple servers in the list
-        * always take the first one
-        */
+       hosts_char = g_strsplit(autoservers, " ", -1);
 
-        hosts_char = g_strsplit(autoservers, " ", -1);
+       good = FALSE;
+       if (ldm.server) {
+           i = 0;
+           while(1) {
+               if(hosts_char[i] == NULL) {
+                   break;
+               }
+               if(!g_strcmp0(hosts_char[i], ldm.server)) {
+                   good = TRUE;
+                   break;
+               }
+               i++;
+           }
+       }
 
-        if ( hosts_char[0] != NULL ) {
-            ldm.server = g_strdup(hosts_char[0]);
-        }
-        return;
+       if (good == FALSE) {
+           ldm.server = g_strdup(hosts_char[0]);
+       }
+       g_strfreev(hosts_char);
+       g_free(autoservers);
+       return;
 }
 
+
 /*
  * mainline
  */
@@ -343,7 +364,24 @@ main(int argc, char *argv[])
     /* decls */
     gchar *display_env, *server_env, *socket_env, *user_env;
     gchar *err_msg = NULL;
+    gchar *greeter_path;
+    struct sigaction action;
 
+    /*
+     * Open log.  Will log locally or back to the server based upon
+     * the boolean LDM_SYSLOG.
+     */
+
+    if (ldm_getenv_bool("LDM_SYSLOG")) {
+        openlog("ldm", LOG_PID | LOG_NOWAIT, LOG_DAEMON);
+    } else {
+        logfile = fopen(LOGFILE, "a");
+        if (!logfile) {
+            fprintf(stderr, "Couldn't open logfile " LOGFILE "\n");
+            exit(1);
+        }
+        setbuf(logfile, NULL);  /* unbuffered writes to the log file */
+    }
 
 #ifdef ENABLE_NLS
     setlocale (LC_ALL, "");
@@ -354,59 +392,27 @@ main(int argc, char *argv[])
 
     g_type_init();
 
+    /* set an empty handler for SIGCHLD so they're not ignored */
+    action.sa_handler = handle_sigchld;
+    sigemptyset(&action.sa_mask);
+    action.sa_flags = SA_SIGINFO;
+    sigaction (SIGCHLD, &action, NULL);
+
     /*
      * Zero out our info struct.
      */
 
     bzero(&ldm, sizeof ldm);
 
-    /*
-     * Process command line args.  Need to get our vt, and our display number
-     * Since we don't have any fancy args, we won't bother to do this with
-     * getopt, rather, we'll just do it manually.
-     */
-
-    if (argc < 3) {
-        usage();
-    }
-
-    if (strncmp(argv[1], "vt", 2)) {         /* sanity */
-        usage();
-    }
-    if (*argv[2] != ':') {                   /* more sanity */
-        usage();
-    }
-
-    ldm.vty = g_strdup(argv[1]);
-    ldm.display = g_strdup(argv[2]);
-
-
-    /*
-     * Open our log.  We want both a fd (for ldm_spawn's to log to) and
-     * a FILE * (for nice fprintf's), so open the log with open(), then
-     * use fdopen to obtain a FILE *
-     * FIXME: Move to syslog, investigate new syslog interface.
-     */
-
-    if ((logfd = open("/var/log/ldm.log",
-                      O_WRONLY | O_CREAT | O_APPEND, 0644)) < 0) {
-        fprintf(stderr, "Couldn't open logfile\n");
-        exit(1);
-    }
-
-    if (!(ldmlog = fdopen(logfd, "a"))) {
-        fprintf(stderr, _("Couldn't fdopen ldmlog\n"));
-        exit(1);
-    }
-
-    setbuf(ldmlog, NULL);      /* Unbuffered */
+    ldm.pid = getpid();  /* Get our pid, to use in the command_socket */
+    ldm.display = g_strdup(getenv("DISPLAY"));
 
     /*
      * Get our IP address.
      */
 
     get_ipaddr();
-    fprintf(ldmlog, _("LDM2 running on ip address %s\n"), ldm.ipaddr);
+    loginfo(_("LDM2 running on ip address %s"), ldm.ipaddr);
 
     /*
      *  Put ip address in environment so that it is available to to the greeter
@@ -418,6 +424,7 @@ main(int argc, char *argv[])
      * Get some of the environment variables we'll need.
      */
 
+    ldminfo_init(&ldminfo_hash, &host_list, getenv("LDM_SERVER"));
     ldm.allowguest = ldm_getenv_bool("LDM_GUESTLOGIN");
     ldm.sound = ldm_getenv_bool("SOUND");
     ldm.sound_daemon = g_strdup(getenv("SOUND_DAEMON"));
@@ -427,34 +434,30 @@ main(int argc, char *argv[])
     ldm.autologin = ldm_getenv_bool("LDM_AUTOLOGIN");
     ldm.lang = g_strdup(getenv("LDM_LANGUAGE"));
     ldm.session = g_strdup(getenv("LDM_SESSION"));
-    ldm.xsession = g_strdup(getenv("LDM_XSESSION"));
-    ldm.greeter_prog = g_strdup(getenv("LDM_GREETER"));
-    if (!ldm.greeter_prog) {
-        ldm.greeter_prog = g_strdup(LDM_EXEC_DIR "/ldmgtkgreet");
+    ldm.sshoptions = g_strdup(getenv("LDM_SSHOPTIONS"));
+    greeter_path = g_strdup(getenv("LDM_GREETER"));
+    if (!greeter_path) {
+        greeter_path = g_strdup("ldmgtkgreet");
+    }
+    if(greeter_path[0] != '/') {
+        ldm.greeter_prog = g_strjoin("/", LDM_EXEC_DIR, greeter_path, NULL);
+        g_free(greeter_path);
+    } else {
+        ldm.greeter_prog = greeter_path;
     }
     ldm.authfile = g_strdup(getenv("XAUTHORITY"));
 
-    display_env = g_strconcat("DISPLAY=", ldm.display, NULL);
-
-    /*
-     * Update our environment with a few extra variables.
-     */
-
-    putenv(display_env);
 
     /*
      * Begin running display manager
      */
 
-    fprintf(ldmlog, _("Launching Xorg\n"));
-    launch_x(argv);         /* Launch X, pass any command line args */
-
     rc_files("init");                      /* Execute any rc files */
 
     if (!ldm.autologin) {
         gint rfd, wfd;
-        fprintf(ldmlog, _("Spawning greeter: %s\n"), ldm.greeter_prog);
-        ldm.greeterpid = ldm_spawn(ldm.greeter_prog, &rfd, &wfd, NULL, NULL);
+        loginfo(_("Spawning greeter: %s"), ldm.greeter_prog);
+        ldm.greeterpid = ldm_spawn(ldm.greeter_prog, &rfd, &wfd, NULL);
         /* create GIOChannels for the greeter */
         ldm.greeterr = g_io_channel_unix_new(rfd);
         ldm.greeterw = g_io_channel_unix_new(wfd);
@@ -464,11 +467,12 @@ main(int argc, char *argv[])
             die(_("ERROR: get_userid from greeter failed"));
         }
 
-	if(!ldm.autologin) {
-	    if (get_passwd()) {
-		die(_("ERROR: get_passwd from greeter failed"));
-	    }
-	}
+        /* If user clicks on guest button above, this has changed */
+        if (!ldm.autologin) {
+            if (get_passwd()) {
+                die(_("ERROR: get_passwd from greeter failed"));
+            }
+        }
 
         if (get_host()) {
             die(_("ERROR: get_host from greeter failed"));
@@ -500,12 +504,9 @@ main(int argc, char *argv[])
     if (!ldm.session) {
         ldm.session = g_strdup("default");
     }
-    if (!ldm.xsession) {
-        err_msg = g_strconcat(_("ERROR: no Xsession"), "\n", NULL );
-    }
 
     if (err_msg) {
-        fprintf(ldmlog, "%s", err_msg);
+        loginfo("%s", err_msg);
         die(_("Fatal error, missing mandatory information"));
     }
 
@@ -514,8 +515,8 @@ main(int argc, char *argv[])
      * control sockets.
      */
 
-    ldm.control_socket = g_strconcat("/var/run/ldm_socket_",
-            ldm.vty, "_", ldm.server, NULL);
+    ldm.control_socket = g_strdup_printf("/var/run/ldm_socket_%d_%s",
+            ldm.pid, ldm.server);
     socket_env = g_strconcat("LDM_SOCKET=", ldm.control_socket, NULL);
     server_env = g_strconcat("LDM_SERVER=", ldm.server, NULL);
     user_env = g_strconcat("LDM_USERNAME=", ldm.username, NULL);
@@ -523,43 +524,36 @@ main(int argc, char *argv[])
     putenv(server_env);
     putenv(user_env);
 
-
     ssh_session();                          /* Log in via ssh */
 
     if (ldm.greeterpid) {
         close_greeter();
     }
 
-    /*
-     * Clear out the password so it's not sitting in memory anywhere
-     */
+    loginfo(_("Established ssh session."));
 
-    g_free(ldm.password);
-
-    fprintf(ldmlog, _("Established ssh session.\n"));
-
-    fprintf(ldmlog, _("Executing rc files.\n"));
+    loginfo(_("Executing rc files."));
     rc_files("start");                      /* Execute any rc files */
-    fprintf(ldmlog, _("Beginning X session.\n"));
+    loginfo(_("Beginning X session."));
     x_session();                            /* Start X session up */
-    fprintf(ldmlog, _("X session ended.\n"));
+    loginfo(_("X session ended."));
 
     /* x_session's exited.  So, clean up. */
 
-    fprintf(ldmlog, _("Executing rc files.\n"));
+    loginfo(_("Executing rc files."));
     rc_files("stop");                       /* Execute any rc files */
 
-    fprintf(ldmlog, _("Killing X server.\n"));
-    kill(ldm.xserverpid, SIGTERM);      /* Kill Xorg server off */
-    ldm_wait(ldm.xserverpid);
-
-    fprintf(ldmlog, _("Ending ssh session.\n"));
+    loginfo(_("Ending ssh session."));
     ssh_endsession();                       /* Log out of server */
 
-    fclose(ldmlog);
-    close(logfd);
+    /* Stop logging */
+    if (logfile) {
+        fclose(logfile);
+    } else {
+        closelog();
+    }
+
     g_free(ldm.server);
-    g_free(ldm.vty);
     g_free(ldm.display);
     g_free(ldm.override_port);
     g_free(ldm.authfile);
@@ -575,6 +569,8 @@ main(int argc, char *argv[])
     g_free(socket_env);
     g_free(server_env);
     g_free(user_env);
-
+    g_list_foreach(host_list, (GFunc)g_free, NULL);
+    g_list_free(host_list);
+    g_hash_table_destroy(ldminfo_hash);
     exit(0);
 }
