@@ -3,9 +3,35 @@
  * LTSP display manager.
  * Manages spawning a session to a server.
  *
- * (c) Scott Balneaves, sbalneav@ltsp.org
+ * Copyright 2007 Scott Balneaves, sbalneav@ltsp.org
  *
  * This software is licensed under the GPL v2 or later.
+
+ * Author: Scott Balneaves, sbalneav@ltsp.org
+
+ * 2007, Scott Balneaves <sbalneav@ltsp.org>
+ * 2008, Scott Balneaves <sbalneav@ltsp.org>
+ *       Ryan Niebur <RyanRyan52@gmail.com>
+ *       Warren Togami <wtogami@redhat.com>
+ *       Vagrant Cascadian <vagrant@freegeek.org>
+ *       Toshio Kuratomi
+
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
+ * published by the Free Software Foundation; either version 2 of the
+ * License, or (at your option) any later version.
+
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+
+ * You should have received a copy of the GNU General Public License
+ * along with this program.  If not, you can find it on the World Wide
+ * Web at http://www.gnu.org/copyleft/gpl.html, or write to the Free
+ * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
+ * MA 02110-1301, USA.
+
  */
 
 #include <stdio.h>
@@ -64,24 +90,34 @@ expect(int fd, char *p, int seconds, ...)
      * Set our file descriptor to be watched.
      */
 
-    FD_ZERO(&set);
-    FD_SET(fd, &set);
-
-    timeout.tv_sec = (long)seconds;             /* one second timeout */
-    timeout.tv_usec = 0;
 
     /*
      * Main loop.
      */
 
     while(1) {
+        timeout.tv_sec = (long)1;             /* one second timeout */
+        timeout.tv_usec = 0;
+
+        FD_ZERO(&set);
+        FD_SET(fd, &set);
         st = select(FD_SETSIZE, &set, NULL, NULL, &timeout);
 
+        if (child_exited) {
+            break;                  /* someone died on us */
+        }
+
         if (st < 0) {                 /* bad thing */
             break;
         }
+
+        if (loopcount == 0) {
+            break;
+        }
+
         if (!st) {                  /* timeout */
-            break;              /* We've not seen the data we want */
+            loopcount--;            /* We've not seen the data we want */
+            continue;
         }
 
         size = read(fd, buffer, sizeof buffer);
@@ -110,12 +146,12 @@ expect(int fd, char *p, int seconds, ...)
         }
     }
 
-    fprintf(ldmlog, "expect saw: %s\n", p);
+    loginfo(_("expect saw: %s"), p);
 
-    if (size < 0 || st < 0) {
+    if (size < 0 || st < 0 || child_exited) {
         return ERROR;               /* error occured */
     }
-    if (st == 0) {
+    if (loopcount == 0) {
         return TIMED_OUT;           /* timed out */
     } else {
         return i;                   /* which expect did we see? */
@@ -135,6 +171,8 @@ ssh_chat(gint fd)
      * password expiry is in the works, then subsequent password prompts
      * will cause us to go back to the greeter. */
 
+    child_exited = FALSE;
+
     while (TRUE) {
         /* ASSUMPTION: ssh will send out a string that ends in ": " for an expiry */
         seen = expect(fd, lastseen, 30, SENTINEL, ": ", NULL);
@@ -142,7 +180,7 @@ ssh_chat(gint fd)
         /* We might have a : in the data, we're looking for :'s at the
            end of the line */
         if (seen == 0) {
-            fprintf(ldmlog, _("Logged in successfully.\n"));
+            loginfo(_("Logged in successfully.\n"));
             g_free(ldm.password);
             ldm.password = NULL;
             return;
@@ -182,12 +220,21 @@ ssh_chat(gint fd)
  */
 
 void
-ssh_session()
+ssh_tty_init(void)
+{
+    (void) setsid();
+    if (login_tty(ldm.sshslavefd) < 0 ) {
+        logerr(_("login_tty failed"));
+        exit(1);
+    }
+}
+
+void
+ssh_session(void)
 {
-    gint fd_master, fd_slave;
     gchar *userathost;
+    gchar *command;
     gchar *port = NULL;
-    GPid pid;
 
     if (ldm.override_port) {
         port = g_strconcat(" -p ", ldm.override_port, " ",  NULL);
@@ -195,65 +242,43 @@ ssh_session()
 
     userathost = g_strconcat(ldm.username, "@", ldm.server, NULL);
 
-    openpty(&fd_master, &fd_slave, NULL, NULL, NULL);
-    ldm.sshfd = fd_master;
+    openpty(&ldm.sshfd, &ldm.sshslavefd, NULL, NULL, NULL);
 
-    /*
-     * Why is g_spawn_async misbehaving?  Back to the old ways.
-     */
+    command = g_strjoin(" ", "ssh", "-Y", "-t", "-M", "-S", ldm.control_socket,
+            port ? port : "",
+            "-o", "NumberOfPasswordPrompts=1",
+            userathost, "echo " SENTINEL "; /bin/sh -", NULL);
+    loginfo(_("ssh_session: %s"), command);
 
-    pid = fork();
+    ldm.sshpid = ldm_spawn(command, NULL, NULL, ssh_tty_init);
 
-    if (pid > 0) {
-        /* In the parent */
-        ldm.sshpid = pid;
-        ssh_chat(fd_master);
-    } else {
-        /* child */
-        int i = 0;
-        int j;
-        char *argv[12];
+    ssh_chat(ldm.sshfd);
 
-        close(fd_master);
-        argv[i++] = "ssh";
-        argv[i++] = "-Y";
-        argv[i++] = "-t";
-        argv[i++] = "-M";
-        argv[i++] = "-S";
-        argv[i++] = ldm.control_socket;
-        if (port) {
-            argv[i++] = "-p";
-            argv[i++] = ldm.override_port;
-        }
-        argv[i++] = userathost;
-        argv[i++] = "echo " SENTINEL "; /bin/sh -";
-        argv[i++] = NULL;
-
-        fprintf(ldmlog, "ssh_session: ");
-        for (j = 0; argv[j]; j++) {
-            fprintf(ldmlog, "%s ", argv[j]);
-        }
-        fprintf(ldmlog, "\n");
-        (void) setsid();
-        if (login_tty(fd_slave) < 0 ) {
-            fprintf(ldmlog, "login_tty failed\n");
-        }
-        execvp(*argv, argv);
+    if (ldm.override_port) {
+        g_free(port);
     }
-
-    g_free(port);
     g_free(userathost);
 }
 
-int
-ssh_endsession()
+void
+ssh_endsession(void)
 {
-    char lastseen[MAXEXP];
-    int status;
+    GPid pid;
+    gchar *command;
+    gchar buf[BUFSIZ];
+    struct stat stbuf;
 
-    write(ldm.sshfd, "exit\r\n", 6);
-    expect(ldm.sshfd, lastseen, 15.0, ldm.server, NULL);
-    status = ldm_wait(ldm.sshpid);
-    close (ldm.sshfd);
-    return status;
+    if (!stat(ldm.control_socket, &stbuf)) {
+        /* socket still exists, so we need to shut down the ssh link */
+
+        command = g_strjoin(" ", "ssh", "-S", ldm.control_socket, "-O", "exit", ldm.server, NULL);
+
+        loginfo(_("Shutting down ssh session: %s"), command);
+        pid = ldm_spawn(command, NULL, NULL, NULL);
+        ldm_wait(pid);
+        read(ldm.sshfd, buf, sizeof buf);       /* clear any exit message so ssh can exit cleanly */
+        ldm_wait(ldm.sshpid);
+        ldm.sshpid = 0;
+        g_free(command);
+    }
 }
